@@ -3,10 +3,9 @@ Submodule that handles the standard workflow, which features more automation and
 """
 import numpy as np
 import pandas as pd
-from collections import Counter
+from collections import Counter, defaultdict
 from tqdm import tqdm
 from hover import module_params
-from hover.annotation import CustomTextcatSession, LabelingFunctionSession
 from hover.evaluation import (
     lf_analysis,
     lf_ensemble,
@@ -15,10 +14,7 @@ from hover.evaluation import (
     composite_accuracy,
 )
 from hover.proposal import metric_sort, fitness_diversity
-from hover.utils.prodigy_model import (
-    create_text_vector_net_from_module,
-    TextVectorNet,
-)
+from hover.future.core.neural import create_text_vector_net_from_module, TextVectorNet
 from hover.utils.torch_helper import vector_dataloader, one_hot, label_smoothing
 from hover.workflow.subroutine import (
     lf_string_representation,
@@ -42,6 +38,7 @@ class Dataset(object):
         -a batch-supervised, minimally noisy train set.
     """
 
+    @todo("This class is deprecated and a new version is in hover.future.core.dataset")
     def __init__(
         self,
         train_dictl,
@@ -163,9 +160,9 @@ class LabelingFunctionPopulation(object):
     Group of labeling functions, divided into a 'candidate' pool and an 'incumbent' pool.
     """
 
-    def __init__(self, candidates, incumbents, **kwargs):
-        self.set_candidates(candidates)
-        self.set_incumbents(incumbents)
+    def __init__(self, candidates=[], incumbents=[], **kwargs):
+        self.candidates = candidates
+        self.incumbents = incumbents
         self.set_params(**kwargs)
 
     def set_params(self, min_lf_per_target=10):
@@ -174,11 +171,21 @@ class LabelingFunctionPopulation(object):
         """
         self.min_lf_per_target = min_lf_per_target
 
-    def set_candidates(self, candidates):
+    @property
+    def candidates(self):
+        return self.__candidates
+
+    @property
+    def incumbents(self):
+        return self.__incumbents
+
+    @candidates.setter
+    def candidates(self, candidates):
         self.__candidates = candidates[:]
         self.__statistics_outdated = True
 
-    def set_incumbents(self, incumbents):
+    @incumbents.setter
+    def incumbents(self, incumbents):
         self.__incumbents = incumbents[:]
         self.__statistics_outdated = True
 
@@ -186,40 +193,18 @@ class LabelingFunctionPopulation(object):
         """
         Deduplicate LFs first among incumbents, then between incumbents and candidates.
         """
-        self.set_incumbents(deduplicate_lfs(self.__incumbents))
-        self.set_candidates(
-            deduplicate_lfs(
-                self.__candidates,
-                lookup_set=set([lf_string_representation(_lf) for _lf in self.__incumbents]),
-            )
+        self.incumbents = deduplicate_lfs(self.incumbents)
+        self.candidates = deduplicate_lfs(
+            self.candidates,
+            lookup_set=set([lf_string_representation(_lf) for _lf in self.incumbents]),
         )
 
-    def candidates(self):
-        return self.__candidates[:]
-
-    def incumbents(self):
-        return self.__incumbents[:]
-
     def num_candidates(self):
         return len(self.__candidates)
 
     def num_incumbents(self):
         return len(self.__incumbents)
 
-    def add_candidates(self, candidates):
-        """
-        Manually add candidates.
-        """
-        self.__candidates += candidates
-        self.__statistics_outdated = True
-
-    def add_incumbents(self, incumbents):
-        """
-        Manually add incumbents.
-        """
-        self.__incumbents += incumbents
-        self.__statistics_outdated = True
-
     def statistics(self, keep_keys=["summary", "agreement"]):
         """
         Read-only.
@@ -247,10 +232,14 @@ class LabelingFunctionPopulation(object):
         agreement_key = "L_train" if use_train else "L_dev"
 
         # calculate statistics: LF empirical accuracy, coverage, and agreement
+        temp_df_dev = dataset.dfs["dev"].copy()
+        temp_df_dev[module_params.ENCODED_LABEL_KEY] = temp_df_dev["label"].apply(
+            lambda x: dataset.label_encoder[x]
+        )
         analysis_result = lf_analysis(
             self.__incumbents + self.__candidates,
             df_train=df_train,
-            df_dev=dataset.dfs["dev"],
+            df_dev=temp_df_dev,
             label_column=module_params.ENCODED_LABEL_KEY,
         )
         for _func in [get_composite_accuracy] + custom_functions:
@@ -321,17 +310,16 @@ class Automated(object):
     End-to-end workflow where the only hands-on part is annotation in Prodigy.
     """
 
-    def __init__(self, dataset, lf_population, model_module_name):
+    def __init__(self, dataset, model_module_name):
         """
-        Start an automated workflow.
+        Start an automated workflow with essential components only.
         :param dataset: workflow.Dataset object.
-        :param lf_population: worklow.LabelingFunctionPopulation object.
         :param model_module_name: path to save and load an in-the-loop model.
         """
         self.logger = module_params.default_logger()
         self.dataset = dataset
-        self.lf_population = lf_population
         self.model_module_name = model_module_name
+        self.__addon_flags = defaultdict(lambda: False)
 
     @todo()
     def run(self, iterations=10):
@@ -353,6 +341,8 @@ class Automated(object):
         Launch Prodigy for annotation on the train set.
         This is step 1/2 for updating the dataset.
         """
+        from hover.annotation import CustomTextcatSession
+
         self.session = CustomTextcatSession(
             self.dataset.dictls["train"], "temp", "text"
         )
@@ -389,6 +379,7 @@ class Automated(object):
         """
         Apply Snorkel's weak supervision to the train set and estimate label probabilities.
         """
+        assert self.__addon_flags["LF"], "No labeling function population set up"
         self.lf_population.run_statistics(self.dataset)
         columns_to_take = self.lf_population.num_incumbents()
         statistics = self.lf_population.statistics(keep_keys=["L_train", "L_dev"])
@@ -417,12 +408,12 @@ class Automated(object):
             texts, output_vectors = self.noisy_train_set()
             batch_size = 64 if len(texts) > 1000 else min(16, len(texts))
         elif key in ["dev", "test"]:
-            texts, labels = (
-                self.dataset.dfs[key]["text"].tolist(),
-                np.array(
-                    self.dataset.dfs[key][module_params.ENCODED_LABEL_KEY].tolist()
-                ),
+            labels = (
+                self.dataset.dfs[key]["label"]
+                .apply(lambda x: self.dataset.label_encoder[x])
+                .tolist()
             )
+            texts = self.dataset.dfs[key]["text"].tolist()
             output_vectors = one_hot(labels, num_classes=len(self.dataset.classes))
             if key == "dev":
                 # the dev set is typically small and clean -- use smaller batch size in case of training on it
@@ -446,19 +437,6 @@ class Automated(object):
         )
         return loader
 
-    def model_from_lf(self, **kwargs):
-        """
-        Train a Prodigy-compatible model from the train set based on labeling functions.
-        Can be useful for cold-starting a model when annotated data is limited.
-        """
-        model = create_text_vector_net_from_module(
-            TextVectorNet, self.model_module_name, self.dataset.classes
-        )
-        train_loader = self.get_loader("train", model.vectorizer, smoothing_coeff=0.01)
-        dev_loader = self.get_loader("dev", model.vectorizer, smoothing_coeff=0.1)
-        train_info = model.train(train_loader, dev_loader, **kwargs)
-        return model, train_info
-
     def model_from_dev(self, **kwargs):
         """
         Train a Prodigy-compatible model from the dev set.
@@ -483,10 +461,13 @@ class Automated(object):
         ).vectorizer
 
         # prepare input vectors to manifold learning
-        subset = ["train", "dev"]
+        subset = ["raw", "train", "dev"]
         fit_texts = []
         for _key in subset:
-            fit_texts += self.dataset.dfs[_key]["text"].tolist()
+            _df = self.dataset.dfs[_key]
+            if _df.empty:
+                continue
+            fit_texts += _df["text"].tolist()
         fit_arr = np.array([vectorizer(_text) for _text in tqdm(fit_texts)])
 
         # initialize and fit manifold learning reducer
@@ -502,11 +483,40 @@ class Automated(object):
             _df["y"] = pd.Series(embedding[start_idx : (start_idx + _length), 1])
             start_idx += _length
 
+    def lf_activate(self, lf_population=None, *args, **kwargs):
+        """
+        Set up the labeling-function add-on utility.
+        :param lf_population: workflow.LabelingFunctionPopulation object.
+        """
+        if lf_population:
+            assert isinstance(
+                lf_population, LabelingFunctionPopulation
+            ), f"Expected LabelingFunctionPopulation, got {type(lf_population)}"
+            self.lf_population = lf_population
+        else:
+            self.lf_population = LabelingFunctionPopulation(*args, **kwargs)
+        self.__addon_flags["LF"] = True
+
+    def model_from_lf(self, **kwargs):
+        """
+        Train a Prodigy-compatible model from the train set based on labeling functions.
+        Can be useful for cold-starting a model when annotated data is limited.
+        """
+        assert self.__addon_flags["LF"], "No labeling function population set up"
+        model = create_text_vector_net_from_module(
+            TextVectorNet, self.model_module_name, self.dataset.classes
+        )
+        train_loader = self.get_loader("train", model.vectorizer, smoothing_coeff=0.01)
+        dev_loader = self.get_loader("dev", model.vectorizer, smoothing_coeff=0.1)
+        train_info = model.train(train_loader, dev_loader, **kwargs)
+        return model, train_info
+
     def lf_rerank(self):
         """
         Score, split, and reorganize LF incumbents and candidates.
         """
 
+        assert self.__addon_flags["LF"], "No labeling function population set up"
         self.logger.divider("Re-ranking labeling functions")
         prev_num_incu, prev_num_cand = (
             self.lf_population.num_incumbents(),
@@ -519,8 +529,8 @@ class Automated(object):
             default_candidate_ranking_function
         )
 
-        self.lf_population.set_incumbents(incu_to_incu + cand_to_incu)
-        self.lf_population.set_candidates(incu_to_cand + cand_to_cand)
+        self.lf_population.incumbents = incu_to_incu + cand_to_incu
+        self.lf_population.candidates = incu_to_cand + cand_to_cand
         num_incu_dup, num_cand_dup = (
             self.lf_population.num_incumbents(),
             self.lf_population.num_candidates(),
@@ -551,15 +561,19 @@ class Automated(object):
         :param autoaccept: whether generated LFs get automatically accepted into candidates, i.e. without manual annotation.
         :type autoaccept: bool
         """
+        from hover.annotation import LabelingFunctionSession
 
+        assert self.__addon_flags["LF"], "No labeling function population set up"
         self.logger.divider("Refilling labeling functions")
         prev_num_cand = self.lf_population.num_candidates()
         raw_candidates = dataset_to_lfs(self.dataset, *args, **kwargs)
 
         # automated analysis that computes composite accuracy and agreement statistics on the dev set
-        analysis_result = lf_analysis(
-            raw_candidates, df_train=None, df_dev=self.dataset.dfs["dev"]
+        temp_df_dev = self.dataset.dfs["dev"].copy()
+        temp_df_dev[module_params.ENCODED_LABEL_KEY] = temp_df_dev["label"].apply(
+            lambda x: self.dataset.label_encoder[x]
         )
+        analysis_result = lf_analysis(raw_candidates, df_train=None, df_dev=temp_df_dev)
         burner_func = compute_on_df({"accuracy": "Emp. Acc.", "coverage": "Coverage"})(
             composite_accuracy
         )
@@ -601,7 +615,7 @@ class Automated(object):
             sess.serve(drop=True)
             accepted_lfs = sess.collect(drop=True)
 
-        self.lf_population.add_candidates(accepted_lfs)
+        self.lf_population.candidates += accepted_lfs
         num_cand = self.lf_population.num_candidates()
         self.logger.good(
             f"Candidates: {prev_num_cand}(prev) + {len(accepted_lfs)}(accepted) = {num_cand}"
